# Copyright 2023 Niklas Birth
#
# This file is part of CoCoPyE.
#
# CoCoPyE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CoCoPyE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CoCoPyE. If not, see <https://www.gnu.org/licenses/>.

import os
import shutil
import subprocess
import sys
import tarfile
import tempfile
import zipfile
from typing import Tuple

from ..external import download, _green, _red, _TICK, _CROSS


def check_uproc(uproc_bin: str) -> Tuple[int, str, str]:
    try:
        process = subprocess.Popen(
            [uproc_bin, '-v'],
            stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, text=True
        )
    except FileNotFoundError:
        return 1, "uproc", _CROSS + " UProC\t\t\t" + _red("not found")

    assert process.stdout is not None

    version = "v" + process.stdout.read().strip().partition("\n")[0].rpartition("version ")[2]

    return 0, "uproc", _TICK + " UProC\t\t\t" + _green(version)


def build_uproc_prot(url: str, install_dir: str, verbose: bool = False) -> None:
    output = subprocess.DEVNULL if not verbose else None

    with tempfile.TemporaryDirectory() as tmpdir:
        download(url, tmpdir, "uproc.tar.gz", "- Downloading UProC repository")
        print("- Downloading UProC repository ✓")

        print("- Extracting repository", end="", flush=True)
        tar = tarfile.open(os.path.join(tmpdir, "uproc.tar.gz"))
        tar.extractall(tmpdir)
        print("\r- Extracting repository ✓")

        print("- Running configure", end="", flush=True)
        configure = subprocess.Popen(["./configure", "--prefix", install_dir],
                                     cwd=os.path.join(tmpdir, "uproc-1.2.0"),
                                     stdout=output, stderr=output)
        if configure.wait() != 0:
            print("\n\nError while running configure. Rerun the command with '--verbose' for subprocess output.")
            sys.exit(1)

        print("\r- Running configure ✓")

        print("- Running make", end="", flush=True)
        make = subprocess.Popen("make", cwd=os.path.join(tmpdir, "uproc-1.2.0"),
                                stdout=output, stderr=output)
        if make.wait() != 0:
            print("\n\nError while running make. Rerun the command with '--verbose' for subprocess output.")
            sys.exit(1)

        print("\r- Running make ✓")

        print("- Running make install", end="", flush=True)
        install = subprocess.Popen(["make", "install"], cwd=os.path.join(tmpdir, "uproc-1.2.0"),
                                   stdout=output, stderr=output)
        if install.wait() != 0:
            print("\n\nError while running make install. Rerun the command with '--verbose' for subprocess output.")
            sys.exit(1)

        print("\r- Running make install ✓\n")


def download_uproc_win(url: str, install_dir: str) -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        download(url, tmpdir, "uproc.zip", "- Downloading UProC")
        print("- Downloading UProC ✓")

        print("- Extracting archive", end="", flush=True)
        with zipfile.ZipFile(os.path.join(tmpdir, "uproc.zip"), 'r') as zip_ref:
            zip_ref.extractall(tmpdir)
        shutil.move(os.path.join(tmpdir, "uproc-1.2.0-win-x86_64"), install_dir)
        print("\r- Extracting archive ✓\n")



# Copyright 2023 Niklas Birth
#
# This file is part of CoCoPyE.
#
# CoCoPyE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CoCoPyE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CoCoPyE. If not, see <https://www.gnu.org/licenses/>.

import _io
import os
import subprocess
from datetime import datetime
from multiprocessing.pool import ThreadPool
from typing import List, Tuple, Dict, Optional

import numpy as np
import numpy.typing as npt
from Bio import SeqIO
from tqdm import tqdm

MAX_PFAM = 17126


def count_pfams(
        orf_bin: str,
        prot_bin: str,
        pfam_dir: str,
        model_dir: str,
        bin_folder: str,
        file_extensions: List[str],
        num_threads: int = 8,
        print_progress: bool = True
) -> Optional[Tuple[npt.NDArray[np.uint8], List[str], List[float]]]:
    """
    This function takes a directory with bins in FASTA format and creates a Pfam count matrix. Each FASTA file is
    considered to be one bin. Sequence headers inside the files (most likely contig ids) are ignored.

    :param orf_bin: Path to uproc-orf binary
    :param prot_bin: Path to uproc-prot binary
    :param pfam_dir: Path to the UProC Pfam database directory
    :param model_dir: Path to UProC model directory
    :param bin_folder: Folder containing input bins in FASTA format
    :param file_extensions: A list of allowed file extensions of the input FASTA files. Probably something like .fna or
    .fasta. Each file in the bin folder that has on of these extensions is considered a bin.
    :param num_threads: Number of threads that UProC should use. It is possible (and likely) that UProC ignores this
    parameter, but you can try.
    :param print_progress: Print a progress bar to stdout
    :return: A 3-tuple. The first element is a QueryMatrix containing the Pfam counts. Each row represents a bin
    and each column a Pfam. The second element is a list of bin names (names of the input FASTA files without file
    extension) in the same order as they appear in the QueryMatrix. The third element is a list of count-ratios of the
    input bins (number of pfams divided by bin size).
    """
    bins = [file for file in os.listdir(bin_folder) if file.rpartition(".")[2] in file_extensions]

    if len(bins) == 0:
        return None

    process_orf = subprocess.Popen(
        orf_bin, stdin=subprocess.PIPE, stdout=subprocess.PIPE, text=True
    )

    assert process_orf.stdin is not None  # MyPy

    process_prot = subprocess.Popen(
        [prot_bin, "-p", "-F", "hf", "-t", str(num_threads), pfam_dir, model_dir],
        stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=process_orf.stdout
    )

    count_pfams_async = ThreadPool(processes=1).apply_async(_count_pfams, (process_prot.stdout,))

    lengths = {}

    for bin_file in tqdm(
            bins,
            ncols=0,
            desc="\033[0;37m[" + str(datetime.now()) + "]\033[0m Counting Pfams",
            disable=not print_progress
    ):
        bin_id = bin_file.rpartition(".")[0]
        for record in SeqIO.parse(os.path.join(bin_folder, bin_file), "fasta"):
            lengths[bin_id] = len(str(record.seq))
            process_orf.stdin.write(">" + bin_id + "$$" + record.id + "\n")
            process_orf.stdin.write(str(record.seq) + "\n")

    process_orf.stdin.close()
    process_orf.wait()

    pfam_counts, sequences, total_counts = count_pfams_async.get()
    count_ratio = [total_counts[seq] / lengths[seq] for seq in sequences]

    result, errors = process_prot.communicate()

    if process_prot.returncode != 0:
        raise Exception(errors)

    return pfam_counts, sequences, count_ratio


def _count_pfams(
        stdout: _io.BufferedReader,
        merge: bool = True
) -> Tuple[npt.NDArray[np.uint8], List[str], Dict[str, int]]:
    pfams: Dict[str, List[int]] = {}
    sequences = []

    for line in iter(stdout.readline, ''):
        line = line.decode("utf-8")
        if line == "":
            break

        seq, pfam = line.split(",")[:2]
        if merge:
            seq = seq.rpartition("$$")[0]
        pfam = int(pfam.strip()[2:])

        if seq not in pfams:
            pfams[seq] = []
            sequences.append(seq)

        pfams[seq].append(pfam)

    count_mat = np.zeros((len(sequences), MAX_PFAM + 1), dtype=np.uint8)
    for idx, seq in enumerate(sequences):
        for pfam in pfams[seq]:
            if count_mat[idx, pfam] == 255:
                continue
            count_mat[idx, pfam] += 1

    total_counts = {key: len(pfams[key]) for key in pfams}

    return count_mat, sequences, total_counts



# Copyright 2023 Niklas Birth
#
# This file is part of CoCoPyE.
#
# CoCoPyE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CoCoPyE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CoCoPyE. If not, see <https://www.gnu.org/licenses/>.

import os
import sys

import argparse
from importlib import resources
from typing import Tuple

import numba
from appdirs import user_config_dir, user_cache_dir, user_log_dir
from tomlkit import parse, dumps, TOMLDocument


ARGS: argparse.Namespace
"""Description"""
CONFIG_FILE: str
"""Description"""
CONFIG: TOMLDocument
"""Description"""


def init() -> None:
    global ARGS, CONFIG_FILE, CONFIG
    CONFIG_FILE, CONFIG = parse_config()
    ARGS = parse_args()


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="cocopye",
        description="Feature-based prediction of genome quality indices",
        epilog="See https://github.com/gobics/cocopye/wiki for more detailed usage instructions."
    )

    parser.add_argument("--pfam24", action='store_true',
                        help="Use Pfam database version 24 (instead of 28)")

    parser.add_argument("--offline", action='store_true',
                        help="Run CoCoPyE in offline mode")

    subparsers = parser.add_subparsers(title="subcommands", dest="subcommand")

    # Subparser run

    run_parser = subparsers.add_parser(
        "run",
        help="Calculate contamination and completeness",
        description="Calculate contamination and completeness"
    )

    run_parser.add_argument("-i", "--infolder", required=True,
                            help="Input folder containing bins in FASTA format")
    run_parser.add_argument("-o", "--outfile", default="cocopye_output.csv",
                            help="Output file (default: cocopye_output.csv)")
    run_parser.add_argument("--file-extensions", default="fasta,fna,fa",
                            help="Allowed file extensions for the FASTA files (default: fasta,fna,fa)")
    run_parser.add_argument("-t", "--threads", default=str(min(8, numba.config.NUMBA_NUM_THREADS)),
                            help="Number of threads")
    run_parser.add_argument("-v", "--verbosity", default="standard",
                            help="Output verbosity (standard, extended, full; default: standard)")
    run_parser.add_argument("--pfam24", action='store_true', dest="pfam24_run",
                            help="Use Pfam database version 24 (instead of 28)")

    # Subparser database

    if CONFIG["advanced"]["enable_db_creator"]:
        db_parser = subparsers.add_parser(
            "database",
            help="Create a new database matrix",
            description="Create a new database matrix"
        )

        db_parser.add_argument("-i", "--infolder", required=True)
        db_parser.add_argument("-m", "--metadata", required=True)
        db_parser.add_argument("-o", "--outfolder", required=True)
        db_parser.add_argument("-t", "--threads", default=str(min(8, numba.config.NUMBA_NUM_THREADS)),
                               help="Number of threads")

    # Subparser toolbox

    setup_parser = subparsers.add_parser(
        "setup",
        help="Tools to setup or repair CoCoPyE",
        description="Tools to setuo or repair CoCoPyE"
    )

    subparsers_toolbox = setup_parser.add_subparsers(title="subcommands", dest="subcommand_setup")

    subparsers_toolbox.add_parser("update-database", help="Update CoCoPyE database")
    subparsers_toolbox.add_parser("cleanup", help="Remove automatically downloaded/generated files")

    testrun_subparser = subparsers_toolbox.add_parser("testrun",
                                                      help="Start a testrun to ensure everything works as expected")
    testrun_subparser.add_argument("--pfam24", action='store_true', dest="pfam24_testrun",
                                   help="Use Pfam database version 24 (instead of 28)")

    dl_subparser = subparsers_toolbox.add_parser("download-dependencies", help="Download missing dependencies")
    dl_subparser.add_argument("-v", "--verbose", action='store_true', help="Show output of subprocesses")
    dl_subparser.add_argument("--pfam24", action='store_true', dest="pfam24_dl",
                              help="Use Pfam database version 24 (instead of 28)")

    # Subparser web

    if CONFIG["advanced"]["enable_webserver"]:
        subparsers.add_parser(
            "web",
            help="Start webserver",
            description="Start webserver"
        )

    return _merge_flag(parser.parse_args(), "pfam24")


def _merge_flag(parsed, flag):
    parsed_dict = vars(parsed)
    keys = [key for key in parsed_dict.keys() if flag in key]

    setattr(parsed, flag, any([parsed_dict[key] for key in keys]))

    for key in keys:
        if key == flag:
            continue
        delattr(parsed, key)

    return parsed


def parse_config() -> Tuple[str, TOMLDocument]:
    # Check several locations for an existing config file
    locations = [
        os.environ.get("COCOPYE_CONFIG", ""),
        "cocopye.toml",
        os.path.join(user_config_dir("cocopye"), "cocopye.toml")
    ]

    for config_file in locations:
        try:
            with open(config_file) as config:
                return config_file, parse(config.read())
                pass
        except IOError:
            pass

    # If we cannot find any, we create a new default configuration file
    print("Unable to find config file.")

    default_config = parse(resources.read_text("cocopye.ui", "config.toml"))
    default_config["server"]["tmpdir"] = os.path.join(user_cache_dir("cocopye"), "server")
    default_config["server"]["logdir"] = os.path.join(user_log_dir("cocopye"), "server")

    os.makedirs(user_config_dir("cocopye"), exist_ok=True)
    f = open(os.path.join(user_config_dir("cocopye"), "cocopye.toml"), "w")
    f.write(dumps(default_config))
    f.close()

    print("Created a new default configuration file at ", os.path.join(user_config_dir("cocopye"), "cocopye.toml.\n"))

    # And then we read it
    try:
        with open(os.path.join(user_config_dir("cocopye"), "cocopye.toml")) as config:
            return os.path.join(user_config_dir("cocopye"), "cocopye.toml"), parse(config.read())
    except IOError:
        print("I wasn't able to read the config file I just created. This shouldn't happen.")
        sys.exit(1)


def change_config(table: str, elem: str, new_value: str) -> None:
    global CONFIG
    CONFIG[table][elem] = new_value

    f = open(CONFIG_FILE, "w")
    f.write(dumps(CONFIG))
    f.close()



"""
This module contains the terminal user interface and the webserver.

### Submodules

- **terminal**: Terminal user interface
- **external**: Functions for checking and downloading external dependencies (binaries and data)
- **config**: Configuration file and command line arguments
- **web**: Webserver
"""


# Copyright 2023 Peter Meinicke, Niklas Birth, Nicolina Leppich
#
# This file is part of CoCoPyE.
#
# CoCoPyE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CoCoPyE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CoCoPyE. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import numpy.typing as npt

_MAX_COUNT = 255


class Histogram:
    """
    A Histogram class that is used for the transformation of a `cocopye.matrices.QueryMatrix` into a
    `cocopye.matrices.FeatureMatrix`.
    """
    _indx_mat: npt.NDArray[np.int32]
    _n_histogram_bins: int

    def __init__(self, resolution: int):
        """
        :param resolution: Histogram resolution. A higher value leads to more histogram bins and therefore a larger
        feature vector.
        """
        self._prepare_histogram(resolution)

    def _prepare_histogram(self, resolution: int) -> None:
        # calcutale borders of bins
        edge_vec = _calc_cr_hist_edges(resolution)
        n_histogram_bins = len(edge_vec) + 1
        # calculate lookup table for ratios
        indx_mat = _calc_index_matrix(edge_vec, n_histogram_bins)

        self._indx_mat = indx_mat
        self._n_histogram_bins = n_histogram_bins

    def calc_bins_for_two_counts(
            self, in_vec: npt.NDArray[np.uint8],
            neighbor_vec: npt.NDArray[np.uint8]
    ) -> npt.NDArray[np.int64]:
        """
        Calculate the bins for two vectors (usually Pfam counts of a query and a neighbor).

        :param in_vec: first (query) vector
        :param neighbor_vec: second (neighbor) vector

        :return: Histogram bins as numpy array
        """
        # clip to maximum count (255 for uint8)!
        # use both count vectors as indices
        # indvec contains all the ratio possitions for both count vect that would be in edge_vec
        ind_vec = self._indx_mat[np.clip(in_vec, 0, _MAX_COUNT), np.clip(neighbor_vec, 0, _MAX_COUNT)]
        # histogram counts for all bins
        x_new_vec = np.bincount(ind_vec[ind_vec != -1], minlength=self._n_histogram_bins)
        return x_new_vec


def _calc_cr_hist_edges(resolution: int) -> npt.NDArray[np.float32]:
    m = resolution + 1  # max. count hyperparameter
    # list of possible count ratios <= 1
    cr_list = [i / j for i in range(1, m) for j in range(i, m)]
    # sorted and non-redundant, now all possible ratios that can occcur
    cr_vec = np.unique(cr_list)

    # build difference to see if two ratios are too close together to form own bins
    diff_vec = cr_vec[1:] - cr_vec[:-1]
    # bulid edges of bins half way between the ratios
    sum_vec = (cr_vec[1:] + cr_vec[:-1]) / 2
    # avoid spurious bins due too numeric "errors"
    edge_half_vec = sum_vec[diff_vec > 1e-8]
    # np.histogram with edges does not provide open intervals for the left/right-most bins
    # we have to add extreme edges for artificial boundaries ...
    # [np.finfo('d').max] ==> Machine limits for floating point types
    # flip takes care of ratios above 1
    edge_vec = np.concatenate(
        (edge_half_vec, np.flip(1 / edge_half_vec)))

    return edge_vec


def _calc_index_matrix(edge_vec: npt.NDArray[np.float32], n_histogram_bins: int) -> npt.NDArray[np.int32]:
    # using 2D index table (256 x 256)!
    indx_mat = np.zeros((_MAX_COUNT + 1, _MAX_COUNT + 1), dtype=np.int32)
    # contains all ratios of 1-255 to 1-255
    # 1/1,  1/2,    1/3     ....    1/255
    # 2/1,  2/2,    2/3     ....    2/255
    # ...
    # 255/1,255/2,  255/3   ....  255/255
    cr_list = [i / j for i in range(1, _MAX_COUNT + 1)
               for j in range(1, _MAX_COUNT + 1)]
    # list of i and j indicies that belong to the same positions in ratios in cr_list
    cr_inds = [(i, j) for i in range(1, _MAX_COUNT + 1)
               for j in range(1, _MAX_COUNT + 1)]
    # tell us where in edge_vec to entries from cr_list need to go
    ind_vec = np.searchsorted(edge_vec, cr_list)
    for ind, val in enumerate(cr_list):
        i, j = cr_inds[ind]
        # tells us vor each count pair i and j with bin belongs to them
        indx_mat[i, j] = ind_vec[ind]

    indx_mat[:, 0] = n_histogram_bins - 1  # last histogram bin (index)
    # 0/0 is not an admissible count ratio -> provoke index error!
    indx_mat[0, 0] = -1
    return indx_mat



# Copyright 2023 Niklas Birth
#
# This file is part of CoCoPyE.
#
# CoCoPyE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CoCoPyE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CoCoPyE. If not, see <https://www.gnu.org/licenses/>.

import gzip
import os
import shutil
import subprocess
import sys
import tarfile
import tempfile
import zipfile
import requests
from typing import Tuple, List, Optional

from appdirs import user_cache_dir, user_data_dir
from packaging.version import Version

from ..external import download, _red, _yellow, _green, _TICK, _CROSS
from ... import constants


def check_pfam_db(pfam_dir: str, version: str = "28") -> Tuple[int, str, str]:
    result = _check_folder(os.path.join(pfam_dir, version),
                           ["fwd.ecurve", "idmap", "prot_thresh_e2", "prot_thresh_e3", "rev.ecurve"])

    if result == "found":
        return 0, "pfam", "  " + _TICK + " Pfam database\t" + _green("v" + version)
    else:
        return 1 if result == "not found" else 2, "pfam", "  " + _CROSS + " Pfam database\t" + _red(result)


def check_model(model_dir: str) -> Tuple[int, str, str]:
    result = _check_folder(model_dir,
                           ["aa_probs", "alphabet", "codon_scores", "orf_thresh_e1", "orf_thresh_e2", "substmat"])

    if result == "found":
        return 0, "model", "  " + _TICK + " Models\t\t" + _green(result)
    else:
        return 1 if result == "not found" else 2, "model", "  " + _CROSS + " Models\t\t" + _red(result)


def check_cocopye_db(db_dir: str, offline: bool = False) -> Tuple[int, str, str]:
    db_files = [
        "universal_Bacteria.npy",
        "universal_Archaea.npy",
        "metadata.csv",
        "count_matrix.npz",
        "model_comp.pickle",
        "model_cont.pickle"
    ]

    result_28 = _check_folder(os.path.join(db_dir, "28"), db_files)
    result_24 = _check_folder(os.path.join(db_dir, "24"), db_files)

    if result_24 == "found" and result_28 == "found":
        version = open(os.path.join(db_dir, "version.txt"), "r").read().strip()
        if not offline and new_db_version_available(version):
            return 0, "cocopye_db", _TICK + " CoCoPyE database\t" + _yellow(version + " (outdated)")
        return 0, "cocopye_db", _TICK + " CoCoPyE database\t" + _green(version)
    else:
        return 1 if result_28 == "not found" else 2, "cocopye_db", _CROSS + " CoCoPyE database\t" + _red(result_28)


def new_db_version_available(current_version):
    try:
        version = requests.get(constants.COCOPYE_DB_LATEST_RELEASE).json()["tag_name"]
    except requests.exceptions.ConnectionError:
        return False
    except KeyError:
        return False

    versions = [version[1:], current_version[1:]]
    versions.sort(key=Version)

    return "v" + versions[-1] != current_version


def _check_folder(folder: str, files: List[str]) -> str:
    try:
        content = os.listdir(folder)
    except FileNotFoundError:
        return "not found"

    return "found" if all([file in content for file in files]) else "error"


def download_pfam_db(url: str, import_bin: str, version: int = 28, verbose: bool = False) -> None:
    output = subprocess.DEVNULL if not verbose else None

    # not using /tmp, because of the large file size
    with tempfile.TemporaryDirectory(prefix="cocopye_", dir=user_cache_dir(None)) as tmpdir:
        download(
            url[version],
            tmpdir,
            "pfam.uprocdb.gz",
            "- Downloading UProC Pfam database"
        )
        print("- Downloading UProc Pfam database ✓")

        print("- Extracting database. This may take a while.", end="", flush=True)
        with gzip.open(os.path.join(tmpdir, "pfam.uprocdb.gz"), "rb") as f_in:
            with open(os.path.join(tmpdir, "pfam.uprocdb"), "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)
        print("\r- Extracting database ✓                                      ")

        print("- Importing database. This may take a while.", end="", flush=True)
        os.makedirs(os.path.join(user_data_dir("cocopye"), "pfam_db", str(version)), exist_ok=True)
        uproc_import = subprocess.Popen([import_bin,
                                         os.path.join(tmpdir, "pfam.uprocdb"),
                                         os.path.join(user_data_dir("cocopye"), "pfam_db", str(version))],
                                        stdout=output, stderr=output)
        if uproc_import.wait() != 0:
            print("\n\nError while running uproc-import. Rerun the command with '--verbose' for subprocess output.")
            sys.exit(1)

        print("\r- Importing database ✓                             \n")


def download_model(url: str) -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        download(
            url,
            tmpdir,
            "model.tar.gz",
            "- Downloading UProC Model"
        )
        print("- Downloading UProc Model ✓")

        print("- Extracting UProC model", end="", flush=True)
        with tarfile.open(os.path.join(tmpdir, "model.tar.gz")) as tar:
            tar.extractall(user_data_dir("cocopye"))
        print("\r- Extracting UProC model ✓\n")


def download_cocopye_db(url: str, db_dir: Optional[str] = None) -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        download(
            url,
            tmpdir,
            "cocopye_db.zip",
            "- Downloading CoCoPyE database"
        )
        print("- Downloading CoCoPyE database ✓")

        print("- Extracting database", end="", flush=True)
        with zipfile.ZipFile(os.path.join(tmpdir, "cocopye_db.zip"), 'r') as zip_ref:
            if db_dir is None:
                db_dir = os.path.join(user_data_dir("cocopye"), "cocopye_db")
            zip_ref.extractall(db_dir)
        print("\r- Extracting database ✓\n")


def update_cocopye_db(url: str, db_dir: str) -> None:
    print("- Removing old database", end="")
    shutil.rmtree(db_dir)
    print("\r- Removing old database ✓")

    download_cocopye_db(url, db_dir)

    print("Database update successful.\n")



# Copyright 2023 Niklas Birth
#
# This file is part of CoCoPyE.
#
# CoCoPyE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CoCoPyE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CoCoPyE. If not, see <https://www.gnu.org/licenses/>.

# === HYPERPARAMETERS ==================================================================================================

K = 4
"""Number of nearest neighbors"""
FRAC_EQ = 1.0
"""Fraction of counts that have to be the same in all neighbors for a Pfam to be considered as a marker"""

RESOLUTION_COMP = {24: 8, 28: 6}
"""Histogram resolution for completeness"""
RESOLUTION_CONT = {24: 7, 28: 12}
"""Histogram resolution for contamination"""

TRANSITION_1_2_MIN_COMP = 0.1
"""Minimal completeness to allow a bin to move from stage 1 to 2"""
TRANSITION_2_3_MIN_COMP = 0.6
"""Minimal completeness to allow a bin to move from stage 2 to 3"""
TRANSITION_2_3_MIN_CONT = 0.3
"""Maximal contamination to allow a bin to move from stage 2 to 3"""

# === DOWNLOAD URLS ====================================================================================================

UPROC = {
    "SRC": "http://uproc.gobics.de/downloads/uproc/uproc-1.2.0.tar.gz",
    "WIN": "http://uproc.gobics.de/downloads/uproc/uproc-1.2.0-win-x86_64.zip"
}
"""
UProC source/binary download path.
Please note that there are some hardcoded directory names in cocopye.ui.external.bin. You might need to adjust them
if you change this URLs (for example in the unlikely case that there will be a UProC update).
"""

UPROC_MODEL = "http://uproc.gobics.de/downloads/models/model.tar.gz"
"""UProC model download path"""

PFAM_DB = {
    24: "http://uproc.gobics.de/downloads/db/pfam24.uprocdb.gz",
    28: "http://uproc.gobics.de/downloads/db/pfam28.uprocdb.gz"
}
"""UProC Pfam database download path"""

COCOPYE_DB = "https://github.com/gobics/cocopye-database/releases/latest/download/database.zip"
"""CocoPyE database download path"""

COCOPYE_DB_LATEST_RELEASE = "https://api.github.com/repos/gobics/cocopye-database/releases/latest"
"""Latest release of the CoCoPyE database (GitHub API). This is used for update checking."""



# Copyright 2023 Niklas Birth
#
# This file is part of CoCoPyE.
#
# CoCoPyE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CoCoPyE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CoCoPyE. If not, see <https://www.gnu.org/licenses/>.

import os
import sys
from datetime import datetime
from typing import List, Dict

import numpy as np
import pandas as pd

from . import constants
from .matrices import DatabaseMatrix, load_u8mat_from_file, QueryMatrix
from .pfam import count_pfams


class Result:
    bin_id: str
    stage: int
    method: str
    count_ratio: float
    knn_scores: float
    taxonomy: str
    taxonomy_level: str
    notes: str

    comp_1_bac: float
    comp_1_arc: float
    cont_1_bac: float
    cont_1_arc: float

    comp_2: float
    cont_2: float
    num_markers_2: int

    comp_3: float
    cont_3: float

    def completeness(self):
        if self.stage == 1:
            return -1
        elif self.stage == 2:
            return self.comp_2
        else:
            return self.comp_3

    def contamination(self) -> float:
        if self.stage == 1:
            return -1
        elif self.stage == 2:
            return self.cont_2
        else:
            return self.cont_3

    def to_csv(self, verbosity: str = "standard") -> str:
        output_lists = {
            "standard": [self.bin_id, "{:.4f}".format(self.completeness()), "{:.4f}".format(self.contamination()), self.method, self.taxonomy,
                         self.taxonomy_level, self.notes],
            "extended": [self.bin_id, "{:.4f}".format(self.completeness()), "{:.4f}".format(self.contamination()), self.stage, self.method, self.num_markers_2,
                         self.count_ratio, self.knn_scores, self.taxonomy, self.taxonomy_level, self.notes],
            "full": [self.bin_id, self.stage, self.method, self.comp_1_arc, self.cont_1_arc, self.comp_1_bac, self.cont_1_bac,
                           self.comp_2, self.cont_2, self.num_markers_2, self.comp_3, self.cont_3, self.count_ratio,
                           self.knn_scores, self.taxonomy, self.taxonomy_level, self.notes]
        }

        return ",".join([str(item) for item in output_lists[verbosity]])

    def to_web(self) -> Dict[str, str]:
        return {
            "completeness": "{:.2f}%".format(self.completeness() * 100),
            "contamination": "{:.2f}%".format(self.contamination() * 100),
            "stage": self.stage,
            "taxonomy": self.taxonomy + " (" + self.taxonomy_level + ")",
        }


def core(cocopye_db: str,
         uproc_orf: str,
         uproc_prot: str,
         pfam_db: str,
         uproc_model: str,
         infolder: str,
         pfam_version: int,
         file_extensions: List[str],
         num_threads: int,
         print_progress: bool = True
         ) -> List[Result]:
    pfam_version = str(pfam_version)

    log("Loading CoCoPyE database", print_progress)
    db_mat = DatabaseMatrix(
        load_u8mat_from_file(os.path.join(cocopye_db, pfam_version, "count_matrix.npz")),
        pd.read_csv(os.path.join(cocopye_db, pfam_version, "metadata.csv"), sep=",")
    )

    pfam_result = count_pfams(
        uproc_orf,
        uproc_prot,
        os.path.join(pfam_db, pfam_version),
        uproc_model,
        infolder,
        file_extensions,
        num_threads,
        print_progress
    )

    if pfam_result is None:
        print("\nError: No input file with extensions " + str(file_extensions) + " found.")
        print("You can use --file-extension to specify a different one. Exiting.")
        sys.exit(1)

    query_mat, bin_ids, count_ratio = pfam_result

    log("Determining nearest neighbors", print_progress)
    query_mat = QueryMatrix(query_mat).with_database(db_mat, constants.K)

    assert len(bin_ids) == query_mat.mat().shape[0]

    log("Calculating preestimates", print_progress)
    universal_arc = np.load(
        os.path.join(cocopye_db, pfam_version, "universal_Archaea.npy"))
    universal_bac = np.load(
        os.path.join(cocopye_db, pfam_version, "universal_Bacteria.npy"))

    preestimates_arc = query_mat.preestimates(universal_arc)
    preestimates_bac = query_mat.preestimates(universal_bac)

    estimates = query_mat.estimates(print_progress=print_progress, frac_eq=constants.FRAC_EQ)

    log("Calculating ML estimates", print_progress)
    feature_mat_comp = query_mat.into_feature_mat(estimates, constants.RESOLUTION_COMP[int(pfam_version)])
    feature_mat_cont = query_mat.into_feature_mat(estimates, constants.RESOLUTION_CONT[int(pfam_version)])

    ml_estimates_comp = feature_mat_comp.ml_estimates(
        os.path.join(cocopye_db, pfam_version, "model_comp.pickle")).clip(0, 1)
    ml_estimates_cont = feature_mat_cont.ml_estimates(
        os.path.join(cocopye_db, pfam_version, "model_cont.pickle")).clip(0, 1000000)

    log("Processing results", print_progress)
    taxonomy = query_mat.taxonomy()
    knn_scores = query_mat.knn_scores()

    completeness = []
    contamination = []
    stage = []
    for idx in range(len(bin_ids)):
        stage.append(1)
        if preestimates_bac[idx, 0] > preestimates_arc[idx, 0]:
            completeness.append(preestimates_bac[idx, 0])
            contamination.append(preestimates_bac[idx, 1])
        else:
            completeness.append(preestimates_arc[idx, 0])
            contamination.append(preestimates_arc[idx, 1])

        if completeness[idx] < constants.TRANSITION_1_2_MIN_COMP:
            continue

        stage[idx] = 2
        completeness[idx] = estimates[idx, 0]
        contamination[idx] = estimates[idx, 1]

        if completeness[idx] < constants.TRANSITION_2_3_MIN_COMP or contamination[idx] > constants.TRANSITION_2_3_MIN_CONT:
            continue

        stage[idx] = 3
        completeness[idx] = ml_estimates_comp[idx]
        contamination[idx] = ml_estimates_cont[idx]

    notes = []
    for idx in range(len(bin_ids)):
        # Currently we do not have any additional notes. But at least we could add some if we want.
        notes.append("")

    results = []
    for idx in range(len(bin_ids)):
        result = Result()

        result.bin_id = bin_ids[idx]
        result.stage = stage[idx]
        result.method = ["rejected", "markers", "markers + neural network"][stage[idx] - 1]
        result.count_ratio = count_ratio[idx]
        result.knn_scores = knn_scores[idx]
        result.taxonomy = taxonomy[idx][0]
        result.taxonomy_level = taxonomy[idx][1]
        result.notes = notes[idx]

        result.comp_1_bac = preestimates_bac[idx, 0]
        result.comp_1_arc = preestimates_arc[idx, 0]
        result.cont_1_bac = preestimates_bac[idx, 1]
        result.cont_1_arc = preestimates_arc[idx, 1]

        result.comp_2 = estimates[idx, 0]
        result.cont_2 = estimates[idx, 1]
        result.num_markers_2 = estimates[idx, 2]

        result.comp_3 = ml_estimates_comp[idx]
        result.cont_3 = ml_estimates_cont[idx]

        results.append(result)

    return results


def log(message: str, show: bool = True):
    """
    Wrapper around the print function for logging purposes.
    """
    if show:
        print("\033[0;37m[" + str(datetime.now()) + "]\033[0m " + message)



# Copyright 2023 Niklas Birth
#
# This file is part of CoCoPyE.
#
# CoCoPyE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CoCoPyE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CoCoPyE. If not, see <https://www.gnu.org/licenses/>.

import os
from typing import Dict

from celery import Celery

from ... import core


app = Celery(
    "tasks",
    backend="redis://localhost",
    broker="redis://localhost",
    # This does not work (crashes on task execution) and I have no idea why.
    # backend=os.getenv("BACKEND"),
    # broker=os.getenv("BROKER"),
    broker_connection_retry=False,
    broker_connection_retry_on_startup=True,
    broker_connection_max_retries=10,
)


@app.task(bind=True, time_limit=os.getenv("CELERY_TIME_LIMIT"))
def estimate_task(self, config, pfam_version: int, infolder: str, debug: bool = False) -> Dict[str, str]:
    self.update_state(state="RUNNING")

    result = core.core(config["external"]["cocopye_db"],
                       config["external"]["uproc_orf_bin"],
                       config["external"]["uproc_prot_bin"],
                       config["external"]["uproc_pfam_db"],
                       config["external"]["uproc_models"],
                       infolder,
                       pfam_version,
                       ["fna"],
                       1,
                       print_progress=debug
                       )[0]

    return result.to_web()



# Copyright 2023 Niklas Birth
#
# This file is part of CoCoPyE.
#
# CoCoPyE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CoCoPyE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CoCoPyE. If not, see <https://www.gnu.org/licenses/>.

"""
This module contains several numba functions that are called by functions in the parent module. In fact, these functions
mostly do the actual work while the matrix classes are just wrappers around them.
"""

from numba import njit, prange
import numpy as np
import numpy.typing as npt
from typing import Tuple, Optional


@njit(nogil=True, parallel=True)
def estimates_njit(query_mat, db_mat, k, frac_eq, progress_bar, knn_inds: npt.NDArray[np.uint64]):
    result = np.zeros((query_mat.shape[0], 3))

    for idx in prange(len(query_mat)):
        comp, cont, num = estimate_njit(db_mat, query_mat[idx], k, frac_eq, knn_inds=knn_inds[idx, :])
        result[idx] = np.array([comp, cont, num])
        progress_bar.update(1)

    return result


@njit
def estimate_njit(
        mat: npt.NDArray[np.uint8],
        vec: npt.NDArray[np.uint8],
        k: int,
        frac_eq: float = 1.0,
        knn_inds: Optional[npt.NDArray[np.uint64]] = None
) -> Tuple[float, float, int]:
    if knn_inds is None:
        knn_mat_idx = nearest_neighbors_idx_njit(mat, vec, k)[0]
        knn_mat = mat[knn_mat_idx, :]
    else:
        knn_mat = mat[knn_inds, :]

    (mode_vals, mode_nums) = mode(knn_mat)

    (mark_inds,) = np.where(
        np.logical_and(np.logical_and(mode_vals > 0, mode_vals < 255), mode_nums >= round(k * frac_eq))
    )
    n_mark = len(mark_inds)

    if n_mark == 0:
        return -1., -1., 0

    comps = np.clip(vec[mark_inds] / mode_vals[mark_inds], 0, 1)
    comp = np.mean(comps)
    cont = np.mean(vec[mark_inds] / mode_vals[mark_inds] - comps)

    return float(comp), float(cont), n_mark


@njit(parallel=True)
def nearest_neighbors_idx_njit_mat(
        db_mat: npt.NDArray[np.uint8],
        q_mat: npt.NDArray[np.uint8],
        k: int
) -> Tuple[npt.NDArray[np.int64], npt.NDArray[np.float32]]:
    knn_inds = np.zeros((q_mat.shape[0], k), dtype=np.uint64)
    knn_scores = np.zeros(q_mat.shape[0], dtype=np.float32)
    for idx in prange(q_mat.shape[0]):
        knn_inds[idx], knn_scores[idx] = nearest_neighbors_idx_njit(db_mat, q_mat[idx], k)
    return knn_inds, knn_scores


@njit(parallel=True)
def nearest_neighbors_idx_njit(
        mat: npt.NDArray[np.uint8],
        vec: npt.NDArray[np.uint8],
        k: int
) -> Tuple[npt.NDArray[np.int64], np.float32]:
    num_refs, num_count = mat.shape

    assert vec.ndim == 1, "Vector has to be 1-dimensional"
    assert vec.shape[0] == num_count, "Vector length must be equal to the number of columns of the matrix"

    eq_counts = np.zeros(num_refs)
    norm_vec = np.sqrt((mat > 0).sum(axis=1))
    for idx in prange(len(mat)):
        eq_count = np.sum(np.logical_and(np.logical_and(0 < vec, vec < 255), mat[idx] == vec))
        eq_counts[idx] = eq_count / norm_vec[idx] / np.sqrt((vec > 0).sum())

    inds = np.flip(np.argsort(eq_counts))[:k]

    return inds, eq_counts[inds].mean()


@njit
def mode(knn_mat: npt.NDArray[np.uint8]):
    mode_vals, mode_nums = np.zeros(knn_mat.shape[1], dtype=np.uint8),  np.zeros(knn_mat.shape[1], dtype=np.uint8)

    for idx, arr in enumerate(knn_mat.T):
        counts = np.bincount(arr)
        mode_vals[idx] = np.argmax(counts)
        mode_nums[idx] = np.max(counts)

    return mode_vals, mode_nums


@njit
def mean_ax0(mat):
    result = np.zeros(mat.shape[1])
    for idx, col in enumerate(mat.T):
        result[idx] = np.mean(col)

    return result


@njit
def std_ax0(mat):
    result = np.zeros(mat.shape[1])
    for idx, col in enumerate(mat.T):
        result[idx] = np.std(col)

    return result
